@@ -11,10 +11,10 @@
 
 module AbsSyn (
   Code, Directive(..), Scheme(..),
-  wrapperName,
+  wrapperCppDefs,
   Scanner(..),
   RECtx(..),
-  RExp(..),
+  RExp(..), nullable,
   DFA(..), State(..), SNum, StartCode, Accept(..),
   RightContext(..), showRCtx, strtype,
   encodeStartCodes, extractActions,
@@ -70,22 +70,23 @@ strtype :: Bool -> String
 strtype True = "ByteString.ByteString"
 strtype False = "String"
 
-wrapperName :: Scheme -> Maybe String
-wrapperName Default {} = Nothing
-wrapperName GScan {} = Just "gscan"
-wrapperName Basic { basicStrType = Str } = Just "basic"
-wrapperName Basic { basicStrType = Lazy } = Just "basic-bytestring"
-wrapperName Basic { basicStrType = Strict } = Just "strict-bytestring"
-wrapperName Posn { posnByteString = False } = Just "posn"
-wrapperName Posn { posnByteString = True } = Just "posn-bytestring"
-wrapperName Monad { monadByteString = False,
-                    monadUserState = False } = Just "monad"
-wrapperName Monad { monadByteString = True,
-                    monadUserState = False } = Just "monad-bytestring"
-wrapperName Monad { monadByteString = False,
-                    monadUserState = True } = Just "monadUserState"
-wrapperName Monad { monadByteString = True,
-                    monadUserState = True } = Just "monadUserState-bytestring"
+
+wrapperCppDefs :: Scheme -> Maybe [String]
+wrapperCppDefs Default {} = Nothing
+wrapperCppDefs GScan {} = Just ["ALEX_GSCAN"]
+wrapperCppDefs Basic { basicStrType = Str } = Just ["ALEX_BASIC"]
+wrapperCppDefs Basic { basicStrType = Lazy } = Just ["ALEX_BASIC_BYTESTRING"]
+wrapperCppDefs Basic { basicStrType = Strict } = Just ["ALEX_STRICT_BYTESTRING"]
+wrapperCppDefs Posn { posnByteString = False } = Just ["ALEX_POSN"]
+wrapperCppDefs Posn { posnByteString = True } = Just ["ALEX_POSN_BYTESTRING"]
+wrapperCppDefs Monad { monadByteString = False,
+                       monadUserState = False } = Just ["ALEX_MONAD"]
+wrapperCppDefs Monad { monadByteString = True,
+                       monadUserState = False } = Just ["ALEX_MONAD_BYTESTRING"]
+wrapperCppDefs Monad { monadByteString = False,
+                       monadUserState = True } = Just ["ALEX_MONAD", "ALEX_MONAD_USER_STATE"]
+wrapperCppDefs Monad { monadByteString = True,
+                       monadUserState = True } = Just ["ALEX_MONAD_BYTESTRING", "ALEX_MONAD_USER_STATE"]
 
 -- TODO: update this comment
 --
@@ -170,6 +171,7 @@ type StartCode = Int
 -- we can generate somewhat faster code in the case that
 -- the lexer doesn't use predicates
 data UsesPreds = UsesPreds | DoesntUsePreds
+  deriving Eq
 
 usesPreds :: DFA s a -> UsesPreds
 usesPreds dfa
@@ -187,7 +189,7 @@ usesPreds dfa
 -- Regular expressions
 
 -- `RExp' provides an abstract syntax for regular expressions.  `Eps' will
--- match empty strings; `Ch p' matches strings containinng a single character
+-- match empty strings; `Ch p' matches strings containing a single character
 -- `c' if `p c' is true; `re1 :%% re2' matches a string if `re1' matches one of
 -- its prefixes and `re2' matches the rest; `re1 :| re2' matches a string if
 -- `re1' or `re2' matches it; `Star re', `Plus re' and `Ques re' can be
@@ -195,13 +197,13 @@ usesPreds dfa
 -- for a formal definition of the semantics of these operators.
 
 data RExp
-  = Eps
-  | Ch CharSet
-  | RExp :%% RExp
-  | RExp :| RExp
-  | Star RExp
-  | Plus RExp
-  | Ques RExp
+  = Eps            -- ^ Empty.
+  | Ch CharSet     -- ^ Singleton.
+  | RExp :%% RExp  -- ^ Sequence.
+  | RExp :| RExp   -- ^ Alternative.
+  | Star RExp      -- ^ Zero or more repetitions.
+  | Plus RExp      -- ^ One  or more repetitions.
+  | Ques RExp      -- ^ Zero or one  repetitions.
 
 instance Show RExp where
   showsPrec _ Eps = showString "()"
@@ -212,6 +214,17 @@ instance Show RExp where
   showsPrec _ (Plus r) = shows r . ('+':)
   showsPrec _ (Ques r) = shows r . ('?':)
 
+-- | A regular expression is nullable if it matches the empty string.
+nullable :: RExp -> Bool
+nullable Eps       = True
+nullable Ch{}      = False
+nullable (l :%% r) = nullable l && nullable r
+nullable (l :|  r) = nullable l || nullable r
+nullable Star{}    = True
+nullable (Plus r)  = nullable r
+nullable Ques{}    = True
+
+
 {------------------------------------------------------------------------------
                           Abstract Regular Expression
 ------------------------------------------------------------------------------}
@@ -320,7 +333,7 @@ encodeStartCodes scan = (scan', 0 : map snd name_code_pairs, sc_hdr)
 -- generated file.
 
 extractActions :: Scheme -> Scanner -> (Scanner,ShowS)
-extractActions scheme scanner = (scanner{scannerTokens = new_tokens}, decl_str)
+extractActions scheme scanner = (scanner{scannerTokens = new_tokens}, decl_str . nl)
  where
   (new_tokens, decls) = unzip (zipWith f (scannerTokens scanner) act_names)
 
@@ -333,64 +346,61 @@ extractActions scheme scanner = (scanner{scannerTokens = new_tokens}, decl_str)
       str "AlexPosn -> Char -> String -> Int -> ((Int, state) -> "
     . str res . str ") -> (Int, state) -> " . str res
 
-  mkDecl fun code = case scheme of
-    Default { defaultTypeInfo = Just (Nothing, actionty) } ->
-        str fun . str " :: " . str actionty . str "\n"
-      . str fun . str " = " . str code . nl
-    Default { defaultTypeInfo = Just (Just tyclasses, actionty) } ->
+  mkDecl  fun code = mkTySig fun
+                   . mkDef fun code
+
+  mkDef   fun code = str fun . str " = " . str code . nl
+
+  mkTySig fun = case scheme of
+    Default { defaultTypeInfo = Just (Nothing, actionty) } -> nl .
+        str fun . str " :: " . str actionty . nl
+    Default { defaultTypeInfo = Just (Just tyclasses, actionty) } -> nl .
       str fun . str " :: (" . str tyclasses . str ") => " .
-      str actionty . str "\n" .
-      str fun . str " = " . str code . nl
-    GScan { gscanTypeInfo = Just (Nothing, tokenty) } ->
-        str fun . str " :: " . gscanActionType tokenty . str "\n"
-      . str fun . str " = " . str code . nl
-    GScan { gscanTypeInfo = Just (Just tyclasses, tokenty) } ->
+      str actionty . nl
+    GScan { gscanTypeInfo = Just (Nothing, tokenty) } -> nl .
+        str fun . str " :: " . gscanActionType tokenty . nl
+    GScan { gscanTypeInfo = Just (Just tyclasses, tokenty) } -> nl .
       str fun . str " :: (" . str tyclasses . str ") => " .
-      gscanActionType tokenty . str "\n" .
-      str fun . str " = " . str code . nl
-    Basic { basicStrType = strty, basicTypeInfo = Just (Nothing, tokenty) } ->
+      gscanActionType tokenty . nl
+    Basic { basicStrType = strty, basicTypeInfo = Just (Nothing, tokenty) } -> nl .
       str fun . str " :: " . str (show strty) . str " -> "
-      . str tokenty . str "\n"
-      . str fun . str " = " . str code . nl
+      . str tokenty . nl
     Basic { basicStrType = strty,
-            basicTypeInfo = Just (Just tyclasses, tokenty) } ->
+            basicTypeInfo = Just (Just tyclasses, tokenty) } -> nl .
       str fun . str " :: (" . str tyclasses . str ") => " .
-      str (show strty) . str " -> " . str tokenty . str "\n" .
-      str fun . str " = " . str code . nl
+      str (show strty) . str " -> " . str tokenty . nl
     Posn { posnByteString = isByteString,
-           posnTypeInfo = Just (Nothing, tokenty) } ->
+           posnTypeInfo = Just (Nothing, tokenty) } -> nl .
       str fun . str " :: AlexPosn -> " . str (strtype isByteString) . str " -> "
-      . str tokenty . str "\n"
-      . str fun . str " = " . str code . nl
+      . str tokenty . nl
     Posn { posnByteString = isByteString,
-           posnTypeInfo = Just (Just tyclasses, tokenty) } ->
+           posnTypeInfo = Just (Just tyclasses, tokenty) } -> nl .
       str fun . str " :: (" . str tyclasses . str ") => AlexPosn -> " .
-      str (strtype isByteString) . str " -> " . str tokenty . str "\n" .
-      str fun . str " = " . str code . nl
+      str (strtype isByteString) . str " -> " . str tokenty . nl
     Monad { monadByteString = isByteString,
-            monadTypeInfo = Just (Nothing, tokenty) } ->
+            monadTypeInfo = Just (Nothing, tokenty) } -> nl .
       let
         actintty = if isByteString then "Int64" else "Int"
       in
         str fun . str " :: AlexInput -> " . str actintty . str " -> Alex ("
-      . str tokenty . str ")\n"
-      . str fun . str " = " . str code . nl
+      . str tokenty . str ")" . nl
     Monad { monadByteString = isByteString,
-            monadTypeInfo = Just (Just tyclasses, tokenty) } ->
+            monadTypeInfo = Just (Just tyclasses, tokenty) } -> nl .
       let
         actintty = if isByteString then "Int64" else "Int"
       in
-        str fun . str " :: (" . str tyclasses . str ") => "
+        str fun . str " :: (" . str tyclasses . str ") =>"
       . str " AlexInput -> " . str actintty
-      . str " -> Alex (" . str tokenty . str ")\n"
-      . str fun . str " = " . str code . nl
-    _ -> str fun . str " = " . str code . nl
+      . str " -> Alex (" . str tokenty . str ")" . nl
+    _ -> id
 
   act_names = map (\n -> "alex_action_" ++ show (n::Int)) [0..]
 
+  decl_str :: ShowS
   decl_str = foldr (.) id [ decl | Just decl <- decls ]
 
 -- -----------------------------------------------------------------------------
 -- Code generation targets
 
 data Target = GhcTarget | HaskellTarget
+  deriving Eq
